<!doctype html>
<title>Luigi Task History</title>
<link href="/static/visualiser/lib/bootstrap3/css/bootstrap.min.css" rel="stylesheet">
<script src="/static/visualiser/lib/bootstrap3/js/bootstrap.min.js"></script>
